#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <signal.h>
#include <netdb.h>
#include <syslog.h>
#include <time.h>
#include "libntp.h"

#define DEFAULT_PORT 23460
#define DNS_LOOK_UP 5*60
#define DNS_RETRY_INTERVAL 60*60
#define dprintf(...) //printf

extern char *optarg;

//static int ntp_enable=0;
static int power_on_delay=30;
static int backoff=15;
static int delay = 1;

typedef struct {
	char server[256];
	short valid;	
} ntp_server;
#include <stdarg.h>
void my_sleep(unsigned int val)
{
	int interval=val;
	while(interval>0)
		interval=sleep(interval);
}

/* Frist is Primary NTP Server  and next is Second NTP Server */
static ntp_server ntp_server_list[]={
	/* GMT - 1, 2, 3 */
	{"time-h.netgear.com", 1},
	// {"204.123.2.5", 1},
	/* GMT - 4, 5, 6 */
	{"time-a.netgear.com", 1},
	/* GMT - 7, 8, 9 */
	{"time-b.netgear.com", 1},
	/* GMT - 10, 11, 12 */
	{"time-c.netgear.com", 1},
	/* GMT + 9, 10, 11 */
	{"time-d.netgear.com", 1},
	/* GMT + 6, 7, 8 */
	{"time-e.netgear.com", 1},
	/* GMT + 3, 4, 5 */
	{"time-f.netgear.com", 1},
	/* GMT + 0, 1, 2 */
	{"time-g.netgear.com", 1},
};

// return 0: DNS ok
// return 1: DNS failed
int try_dns_lookup(char *host)
{
	int i=3; /* 3 times */
	struct hostent *ntpserver=NULL;
	
	while(i-- > 0){
		ntpserver = gethostbyname(host);
		if (ntpserver == NULL) {
			my_sleep(DNS_LOOK_UP);
			continue;
		}
		else if (ntpserver->h_length != 4) {
			my_sleep(DNS_LOOK_UP);
			continue;
		}
		else {
			return 0;
		}
		dprintf("try_dns_lookup: i = %d\n", i);
	}
	dprintf("try_dns_lookup\n");
	return -1;
}
#if 0
void signal_handle(int sign)
{
	dprintf("caught signal!\n");
	switch(sign){
		case SIGUSR1:
			if(ntp_enable==0){
				printf("SIGUSR1\n");
				ntp_enable=1;
				/* power on delay 60 - 300 secs */
				power_on_delay=60+(int) (240.0*rand()/(RAND_MAX+1.0));
				if (delay == 1) {
					my_sleep(power_on_delay);
				}
			}
			break;
		case SIGUSR2:
			ntp_enable=0;
			break;
	}
}
#endif
int check_wan_up()
{
	if(access("/tmp/ntp_start",F_OK)==0){
		/* power on delay 60 - 300 secs */
		/* remove this for netgear */
		//power_on_delay=60+(int) (240.0*rand()/(RAND_MAX+1.0));
		if (delay == 1 && power_on_delay!=0) {
			my_sleep(power_on_delay);
			power_on_delay=0;
		}
		return 1;
	}
	return 0;
}

// parameter: time server to do DNS lookup first
// return: 0 -> time server
//         1 -> time server
//        -1 -> DNS lookup failed on both time servers
int dns_ntp_server(int server0, ntp_server * my_server) {

	int server1 = 0;
	
	if (server0 == 0) {
		server1 = 1;
	} else if (server0 == 1) {
		server1 = 0;
	}

	if (try_dns_lookup(my_server[server0].server) == 0) {
		return server0;
	} else if (try_dns_lookup(my_server[server1].server) == 0) {
		return server1;
	} 
		
	return -1;
}


// request NTP time
// return: NTP_ERR -> no valid NTP packet received
// 	   NTP_KOD -> received a KOD packet
int req_ntp_time(int sockfd, int timeserver, ntp_server * my_server) {
	
	int retval1 = 0;
	int retval2 = 0;
	int retval3 = 0;
	char ntpserver[16];
	char buf[1500]; 
		
	openlog("ntp",0,LOG_SYSLOG);
	syslog(LOG_INFO,"Send out NTP request to %s", my_server[timeserver].server);
	retval1 = send_request(sockfd, my_server[timeserver].server, 123);
	retval2 = recv_response(sockfd,buf,30,ntpserver);
	
	dprintf ("done send and recv!\n");
	if ((retval1 != NTP_OK) || (retval2 != NTP_OK)) {
		closelog();
		return NTP_ERR;
	}
	
	dprintf("now parsing the packet!\n");
	
	retval3 = handle_ntp_reply(buf);
	if (retval3 == NTP_ERR) {
		closelog();
		return NTP_ERR;
	} else if (retval3 == NTP_KOD) {
		closelog();
		return NTP_KOD;
	} else if (retval3 == NTP_OK) {
		syslog(LOG_INFO,"Receive NTP Reply from %s", my_server[timeserver].server);
		closelog();
		return NTP_OK;
	}

	return NTP_ERR;
}	
	
// get the primary and secondary time server 
// according to the time zone
int get_timeserver(ntp_server * my_server, char * gmt_zone) {

	int i = 0;
	int j = 0;
	int zone = 0;
	char * tail;
	
	tail = &gmt_zone[9];
	// parse gmt_zone
	zone = strtol(&gmt_zone[3], &tail, 0);	

		
	switch (zone) {
		case +1:
		case +2:
		case +3: zone = 0;
			 break;
		case +4:
		case +5:
		case +6: zone = 1;
			 break;
		case +7:
		case +8:
		case +9: zone = 2;
			 break;
		case +10:
		case +11:
		case +12: zone = 3;
			  break;
		case -9:
		case -10:
		case -11: zone = 4;
			 break;
		case -6:
		case -7:
		case -8: zone = 5;
			break;
		case -3:
		case -4:
		case -5: zone = 6;
			break;
		case 0:
		case -1:
		case -2: zone = 7;
			break;
		default:
			zone = 0;
	}
			 
	j = zone + 2;
	
	if (zone > 7 || zone < 0) {
		dprintf("Error: invalid time zone %d!\n", zone);
		return -1;
	}
	
	if (my_server[0].valid == 1) {
		i = 1;
	}

	while (i < 2) {
		strcpy(my_server[i].server, ntp_server_list[zone].server);
		my_server[i].valid = ntp_server_list[zone].valid;
		zone++;
		zone = zone % 8;
		i++;
	}

	return 0;
}

int produce_sysinit_log(void)
{
    char *week[]={"Sunday",
                   "Monday",
                   "Tuesday",
                   "Wednesday",
                   "Thursday",
                   "Friday",
                   "Saturday"};
    char *month[]={"Jan","Feb","March","Apr","May","Jun",
                    "July","Aug","Sep","Oct","Nov","Dec"};
    time_t t;
	struct tm *st;
	FILE *fp;
	char tmp[40]={0};
	
    t=time(NULL);
	st=localtime(&t);
	if(st!=NULL){
	    fp=fopen("/var/otherlog","w");
	    if(fp == NULL)
    	{
    	    return -1;
    	}

    	fprintf(fp,"[Initialized] %s, %d %s %d %02d:%02d:%02d\n",
    	        week[st->tm_wday]
    	        ,st->tm_mday
    	        ,month[st->tm_mon]
    	        ,st->tm_year+1900
    	        ,st->tm_hour
    			,st->tm_min
    			,st->tm_sec);

    	fclose(fp);
    }
	
	return 0;
	
}


void ntp_process_flow(int sockfd, ntp_server * my_server)
{
	int retval = 0;
	int timeserver=0;

	dprintf("inside ntp_process_flow()\n");

	while(1){
		
		// this is the upper part DNS lookup 
		// (see the Netgear home router flow chart V1.5)
		while (1) {
			timeserver = dns_ntp_server(0, my_server);
			
			if (check_wan_up() == 0) {
				return;
			}

			if (timeserver == -1) {
				my_sleep(DNS_RETRY_INTERVAL); /* 1 hr */
			} else {
				break;
			}
		}	
	
		dprintf("Got timeserver: %d\n", timeserver);

		/* backoff 15 - 60 secs */
		backoff=15+(int) (45.0*rand()/(RAND_MAX+1.0));

		while (1) {

			// check if NTP client is enabled again
			// if not, then go back to the beginning of the loop
			//if (ntp_enable == 0) {
			if (check_wan_up() == 0) {
				return;
			}
			
			dprintf("check for my_server list\n");
			if (my_server[0].valid == 0 && my_server[1].valid == 0) {
				my_sleep(1);
				continue;
			}
			
			dprintf("calling req_ntp_time()\n");
			retval = req_ntp_time(sockfd, timeserver, my_server);
		
			
			if (retval == NTP_OK) {	
				// sync system clock
				dprintf("Got NTP_OK, now sleep for 72 hrs\n");	
				produce_sysinit_log();
				// sleep 70 hours, then start with dns lookup again
				my_sleep (252000);
				break;
			
			} else if (retval == NTP_KOD) {
				// remove the current timeserver from my_server
				dprintf("Got KOD\n");				
				my_server[timeserver].valid = 0;
			
			} else if (retval == NTP_ERR) {
				// not getting valid NTP packet, do backoff
				dprintf("Got NTP_ERR\n");
				while (backoff <= 252000) {
					dprintf("now sleep for %d seconds!\n", backoff);
					my_sleep (backoff);
					backoff = backoff * 2;
					if (timeserver == 0) {
						timeserver = dns_ntp_server(1, my_server);
						if (timeserver != -1) {
							break;
						}
					} else if (timeserver == 1) {
						timeserver = dns_ntp_server(0, my_server);
						if (timeserver != -1) {
							break;
						}
					}
				}

				if (backoff <= 252000) {
					continue;
				} else {
					break;
				}
			}
		}
	}
}


void test(int zone) {

	return;
}

int main(int argc, char **argv)
{
	int sockfd = -1;
	int i = 0;
	char c;
	char zone[10];

	ntp_server my_server[2];

	
	//signal(SIGUSR1,&signal_handle);
	//signal(SIGUSR2,&signal_handle);

	memset(my_server, 0x00, sizeof(my_server));
	/*
	close(0);
	close(1);
	close(2);
	*/
//	daemon(1, 1);

	for (;;) {
		c = getopt( argc, argv, "h:z:c:d:f:");
		if (c == EOF) break;
		switch (c) {
			case 'h':
				strcpy(my_server[0].server, optarg);
				my_server[0].valid = 1;
				break;
			case 'z':
				strcpy(zone, optarg);
				break;
			case 'c':

				break;
			case 'd':
				delay = 0;
				break;
			case 'f':
				break;

			default:
				exit(1);
		}
	}
	if (get_timeserver(my_server, zone) != 0) {
                dprintf("failed to get time servers!\n");
		return 0;
	}
        while (i < 2) {
		dprintf("%s\n", my_server[i].server);
		i++;
	}
	// this is the beginning of the LOOP
	while (1) {
	
		/* when ntp is disabled, sleep 10 sec */
		//while(ntp_enable == 0)
		while(check_wan_up() == 0){
			dprintf("sleeping...\n");
			my_sleep(10);
		}	
		dprintf("calling setup_ntp_socket()\n");
		while((sockfd=setup_ntp_socket(INADDR_ANY,23460))<0){
				dprintf("sockfd error\n");
				my_sleep(1);
		}
		dprintf("calling ntp_process_flow()\n");
		ntp_process_flow(sockfd, my_server);	
	
		close(sockfd);
	}
	
	return 0;
}
