//=================================================================================
// Topro PC camera definition
//=================================================================================
//#define TP6800_QVGA
#define TP6800_MT9V011      0
#define TP6800_7131E        0
#define TP6800_HV7131R      0
#define TP6800_PAS302       0
#define TP6800_EVS          0
#define TP6800_TAS          0
#define TP6800_SHARPCCD     1

#define MT9V011           1
#define HV7131R           2
#define EVS               3
#define TAS               4
#define PAS302            5
#define SHARPCCD          6

#ifndef TP_DEF_H
#define TP_DEF_H

// Topro read & write command
#define READ_REGISTER   		0x0D
#define WRITE_REGISTER  		0x0E
#define GET_CONFIG_REGISTER		0x08

#define READ_REGISTER_TYPE   		0xC0
#define WRITE_REGISTER_TYPE  		0x40
#define GET_CONFIG_REGISTER_TYPE	0x80

// Topro register definition
#define SIF_TYPE                0x10
#define SIF_CONTROL        	0x11
#define SIF_ADDR_S           	0x12
#define SIF_TX_DATA           	0x13
#define SIF_RX_DATA           	0x14
#define GPIO_PU			0x15
#define GPIO_PD		        0x16
#define GPIO_IO			0x17
#define GPIO_DATA              	0x18
#define SIF_ADDR_S2           	0x19
#define SIF_TX_DATA2           	0x1A
#define SIF_RX_DATA2           	0x1B
#define ENDP_1_CTL              0x21
#define TIMING_CFG              0x2F
#define SENSOR_CFG              0x30
#define PIXEL_START             0x31
#define PIXEL_END_L             0x32
#define PIXEL_END_H             0x33
#define LINE_START              0x34
#define LINE_END_L              0x35
#define LINE_END_H              0x36
#define FRONT_DARK_ST           0x37
#define FRONT_DARK_END          0x38
#define REAR_DARK_ST_L          0x39
#define REAR_DARK_ST_H          0x3A
#define REAR_DARK_END_L         0x3B
#define REAR_DARK_END_H         0x3C
#define HORIZONTAL_DARK_LINE_L  0x3D
#define HORIZONTAL_DARK_LINE_H  0x3E
#define FRAME_RATE              0x3F
#define EXPOSURE_L              0x40
#define EXPOSURE_H              0x41
#define VSYNC_EN_PERIOD_L       0x42
#define VSYNC_EN_PERIOD_H       0x43
#define VSYNC_DIS_PERIOD        0x44
#define HSYNC_EN_PERIOD_L       0x45
#define HSYNC_EN_PERIOD_H       0x46
#define HSYNC_DIS_PERIOD        0x47
#define HSYNC_TO_VSYNC_L        0x48
#define HSYNC_TO_VSYNC_H        0x49

#define BAD_LOW_THRLD_L         0x50
#define BAD_LOW_THRLD_H         0x51
#define BAD_HIGH_THRLD_L        0x52
#define BAD_HIGH_THRLD_H        0x53
#define DARK_CFG                0x54
#define GAMMA_R                 0x55
#define GAMMA_G                 0x56
#define GAMMA_B                 0x57
#define EDGE_THRLD              0x5C
#define DEMOSAIC_CFG            0x5D

#define Y_GAIN_RL               0x60
#define Y_GAIN_RH               0x61
#define Y_GAIN_GL               0x62
#define Y_GAIN_GH               0x63
#define Y_GAIN_BL               0x64
#define Y_GAIN_BH               0x65

#define U_GAIN_RL               0x66
#define U_GAIN_RH               0x67
#define U_GAIN_GL               0x68
#define U_GAIN_GH               0x69
#define U_GAIN_BL               0x6A
#define U_GAIN_BH               0x6B

#define V_GAIN_RL               0x6C
#define V_GAIN_RH               0x6D
#define V_GAIN_GL               0x6E
#define V_GAIN_GH               0x6F
#define V_GAIN_BL               0x70
#define V_GAIN_BH               0x71

#define FORMAT                  0x78
#define QUALITY                 0x79
#define BLK_THRLD               0x7A
#define BUFFER_STATUS           0x7B

#define ACC_R0                  0x80
#define ACC_R1                  0x81
#define ACC_R2                  0x82
#define ACC_R3                  0x83

#define ACC_G0                  0x84
#define ACC_G1                  0x85
#define ACC_G2                  0x86
#define ACC_G3                  0x87

#define ACC_B0                  0x88
#define ACC_B1                  0x88
#define ACC_B2                  0x8A
#define ACC_B3                  0x8B

#define MAX_R                   0x8C
#define MAX_G                   0x8D
#define MAX_B                   0x8E
#define MAX_Y                   0x8F

#define ACC_Y0                  0x90
#define ACC_Y1                  0x91
#define ACC_Y2                  0x92
#define ACC_Y3                  0x93

#define END_REG                 0xFF

struct topro_sregs{
	unsigned char index;
	unsigned char value;
};

#define GAMMA_TABLE_LENGTH      1024
#define GAIN_TABLE_LENGTH       18

#define maximum(x, y)         ((x > y) ? x : y)
#define minimum(x, y)         ((x > y) ? y : x)

void topro_cam_init(struct usb_device *udev);

unsigned char topro_read_reg(struct usb_device *udev, unsigned char index);
int topro_write_reg(struct usb_device *udev, unsigned char index, unsigned char data);
int topro_write_i2c(struct usb_device *udev, unsigned char dev, unsigned char reg, unsigned char data_h, unsigned char data_l);
int topro_prog_regs(struct usb_device *udev, struct topro_sregs *ptpara);

int topro_setting_iso(struct usb_device *udev, unsigned char enset);

int topro_bulk_out(struct usb_device *udev, unsigned char bulkctrl, unsigned char *pbulkdata, int bulklen);
int topro_prog_gamma(struct usb_device *udev, unsigned char bulkctrl);
int topro_prog_gain(struct usb_device *udev, unsigned char *pgain);
void topro_set_gamma(struct usb_device *udev);
void topro_set_hue_saturation(struct usb_device *udev, long hue, long saturation);
void topro_autowhitebalance(struct usb_device *udev);
void topro_autoexposure(struct usb_device *udev);
void topro_setting_sensor(int Gain,int gFrameRate,int Exposure);
void topro_setexposure(struct usb_device *udev, int Exposure);
void CCD_Control(struct usb_device *udev, unsigned char* pData, int Length);


#endif

